"""
Functions to manipulate and analyze voxels.
"""


from copy import deepcopy
import ctypes

import numpy as np
import pandas as pd

from pyntcloud import PyntCloud
from pyntcloud.structures.voxelgrid import VoxelGrid

from volumizer import utils, align
from volumizer.constants import OCCLUDED_DIMENSION_LIMIT, MIN_NUM_VOXELS
from volumizer.types import VoxelGroup
from volumizer.paths import C_CODE_DIR


if utils.using_performant():
    VOXEL_C_PATH = C_CODE_DIR / "voxel.so"
    VOXEL_C = ctypes.CDLL(str(VOXEL_C_PATH.absolute()))


def coords_to_point_cloud(coords: pd.DataFrame) -> PyntCloud:
    """
    Produce a point-cloud from atomic coordinates.
    """
    return PyntCloud(coords)


def add_voxel_grid(cloud: PyntCloud) -> tuple[PyntCloud, str]:
    """
    Generate a voxel grid surrounding the point-cloud.
    """
    voxel_grid_id = cloud.add_structure(
        "voxelgrid",
        size_x=utils.VOXEL_SIZE,
        size_y=utils.VOXEL_SIZE,
        size_z=utils.VOXEL_SIZE,
        regular_bounding_box=False,
    )
    return cloud, voxel_grid_id


def get_voxel_grid(cloud: PyntCloud, voxel_grid_id: str) -> VoxelGrid:
    """
    Generate an array representing a binary voxel grid where zero represents no protein
    atoms in that voxel (e.g. solvent) and a non-zero represents a protein atom in that voxel.
    """
    return cloud.structures[voxel_grid_id]


def get_protein_solvent_voxel_array(voxel_grid: VoxelGrid) -> np.ndarray:
    """
    Generate a 3D array of x[y[z]] as the voxel coordinate and the value of that array entry
    being zero for no protein and one when a protein atom is in that voxel.
    """
    return voxel_grid.get_feature_vector(mode="binary")


def get_protein_and_solvent_voxels(
    binary_voxel_array: np.ndarray,
    voxel_grid_dimensions: np.ndarray,
) -> tuple[VoxelGroup, VoxelGroup]:
    """
    From the voxel grid, return the indices of the protein containing voxels,
    and those not containing protein (e.g. solvent).
    """
    protein_voxels = np.nonzero(binary_voxel_array)
    solvent_voxels = np.nonzero(binary_voxel_array == 0)

    protein_voxel_indices = compute_voxel_indices(protein_voxels, voxel_grid_dimensions)
    solvent_voxel_indices = compute_voxel_indices(solvent_voxels, voxel_grid_dimensions)

    return (
        VoxelGroup(
            voxels=protein_voxels,
            indices=protein_voxel_indices,
            num_voxels=len(protein_voxel_indices),
            voxel_type="protein",
            volume=compute_voxel_group_volume(len(protein_voxel_indices)),
        ),
        VoxelGroup(
            voxels=solvent_voxels,
            indices=solvent_voxel_indices,
            num_voxels=len(solvent_voxel_indices),
            voxel_type="solvent",
            volume=compute_voxel_group_volume(len(solvent_voxel_indices)),
        ),
    )


def get_occluded_dimensions(
    query_voxel: tuple[int, int, int],
    occluding_x: list[int],
    occluding_y: list[int],
    occluding_z: list[int],
) -> list[int]:
    """
    Determine how many ordinal axes are occluded.

    We simply compare our query voxel with the coordinates of the `occluding` voxels,
    which are e.g. protein voxels.

    For performance reasons rather than looking across ALL occluding voxels, we look at planar arrays
    of coordinates.
    """
    occluded_dimensions = [0, 0, 0, 0, 0, 0]

    # here we assume that e.g. occluding_z[0] is the min value and occluding_z[-1] is the max
    #   if this becomes untrue we would need to explicitly take the min/max
    #   this is less performant for large structures
    if len(occluding_z) != 0:
        if occluding_z[0] < query_voxel[2]:
            occluded_dimensions[4] = 1
        if occluding_z[-1] > query_voxel[2]:
            occluded_dimensions[5] = 1
    if len(occluding_y) != 0:
        if occluding_y[0] < query_voxel[1]:
            occluded_dimensions[2] = 1
        if occluding_y[-1] > query_voxel[1]:
            occluded_dimensions[3] = 1
    if len(occluding_x) != 0:
        if occluding_x[0] < query_voxel[0]:
            occluded_dimensions[0] = 1
        if occluding_x[-1] > query_voxel[0]:
            occluded_dimensions[1] = 1

    return occluded_dimensions


def is_buried(occluded_dimensions: list[int]) -> bool:
    """
    If 5 or 6 dimensions are occluded, return True.
    If less than 4 dimensions are occluded, return False.
    If exactly 4 dimensions are occluded, return True if the two unoccluded dimensions are the same axis,
    False otherwise
    """

    if occluded_dimensions.count(1) > OCCLUDED_DIMENSION_LIMIT:
        return True
    elif occluded_dimensions.count(1) == OCCLUDED_DIMENSION_LIMIT:
        if occluded_dimensions[0] == 0 and occluded_dimensions[1] == 0:
            return True
        elif occluded_dimensions[2] == 0 and occluded_dimensions[3] == 0:
            return True
        elif occluded_dimensions[4] == 0 and occluded_dimensions[5] == 0:
            return True

    return False


def build_planar_voxel_coordinate_arrays(
    voxels: tuple[np.ndarray, ...], voxel_grid_dimensions: np.ndarray
) -> tuple[list[list[list[int]]], list[list[list[int]]], list[list[list[int]]],]:
    """
    For each two-dimension pair, construct a tuple of the min/max coordinates in the 3rd dimension.
    """
    z_array = [
        [[] for y in range(voxel_grid_dimensions[1])]
        for x in range(voxel_grid_dimensions[0])
    ]
    y_array = [
        [[] for z in range(voxel_grid_dimensions[2])]
        for x in range(voxel_grid_dimensions[0])
    ]
    x_array = [
        [[] for z in range(voxel_grid_dimensions[2])]
        for y in range(voxel_grid_dimensions[1])
    ]

    for i in range(voxels[0].size):
        z_array[voxels[0][i]][voxels[1][i]].append(voxels[2][i])
        y_array[voxels[0][i]][voxels[2][i]].append(voxels[1][i])
        x_array[voxels[1][i]][voxels[2][i]].append(voxels[0][i])

    return z_array, y_array, x_array


def get_exposed_and_buried_voxels(
    solvent_voxels: VoxelGroup,
    protein_voxels: VoxelGroup,
    voxel_grid_dimensions: np.ndarray,
) -> tuple[VoxelGroup, VoxelGroup]:
    """
    Use simple geometric heuristics to determine if a group of solvent voxels is buried or exposed.
    """
    buried_voxels = ([], [], [])
    exposed_voxels = ([], [], [])

    z_array, y_array, x_array = build_planar_voxel_coordinate_arrays(
        protein_voxels.voxels, voxel_grid_dimensions
    )

    for i in range(solvent_voxels.voxels[0].size):
        query_voxel = (
            solvent_voxels.voxels[0][i],
            solvent_voxels.voxels[1][i],
            solvent_voxels.voxels[2][i],
        )

        occluded_dimensions = get_occluded_dimensions(
            query_voxel,
            x_array[query_voxel[1]][query_voxel[2]],
            y_array[query_voxel[0]][query_voxel[2]],
            z_array[query_voxel[0]][query_voxel[1]],
        )

        if is_buried(occluded_dimensions):
            buried_voxels[0].append(query_voxel[0])
            buried_voxels[1].append(query_voxel[1])
            buried_voxels[2].append(query_voxel[2])
        else:
            exposed_voxels[0].append(query_voxel[0])
            exposed_voxels[1].append(query_voxel[1])
            exposed_voxels[2].append(query_voxel[2])

    buried_voxel_indices = compute_voxel_indices(buried_voxels, voxel_grid_dimensions)
    exposed_voxel_indices = compute_voxel_indices(exposed_voxels, voxel_grid_dimensions)

    return (
        VoxelGroup(
            voxels=(
                np.array(exposed_voxels[0]),
                np.array(exposed_voxels[1]),
                np.array(exposed_voxels[2]),
            ),
            indices=exposed_voxel_indices,
            num_voxels=len(exposed_voxel_indices),
            voxel_type="exposed",
            volume=compute_voxel_group_volume(len(exposed_voxel_indices)),
        ),
        VoxelGroup(
            voxels=(
                np.array(buried_voxels[0]),
                np.array(buried_voxels[1]),
                np.array(buried_voxels[2]),
            ),
            indices=buried_voxel_indices,
            num_voxels=len(buried_voxel_indices),
            voxel_type="buried",
            volume=compute_voxel_group_volume(len(buried_voxel_indices)),
        ),
    )


def get_neighbor_voxels_python(
    query_voxels: tuple[np.ndarray, np.ndarray, np.ndarray],
    reference_voxels: tuple[np.ndarray, np.ndarray, np.ndarray],
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Return the voxels from a query group that neighbor a reference group.
    """
    neighbor_indices = []
    for query_index in range(len(query_voxels[0])):
        query_voxel = get_single_voxel(query_voxels, query_index)
        for reference_index in range(len(reference_voxels[0])):
            reference_voxel = get_single_voxel(reference_voxels, reference_index)
            if is_neighbor_voxel(query_voxel, reference_voxel):
                neighbor_indices.append(query_index)
                break

    return (
        np.array([query_voxels[0][i] for i in neighbor_indices]),
        np.array([query_voxels[1][i] for i in neighbor_indices]),
        np.array([query_voxels[2][i] for i in neighbor_indices]),
    )


def get_neighbor_voxels_c(
    query_voxels: tuple[np.ndarray, np.ndarray, np.ndarray],
    reference_voxels: tuple[np.ndarray, np.ndarray, np.ndarray],
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Return the voxels from a query group that neighbor a reference group.
    """
    # make the voxels compatible with C
    num_query = len(query_voxels[0])
    query_x = (ctypes.c_int * num_query)(*query_voxels[0])
    query_y = (ctypes.c_int * num_query)(*query_voxels[1])
    query_z = (ctypes.c_int * num_query)(*query_voxels[2])

    num_reference = len(reference_voxels[0])
    reference_x = (ctypes.c_int * num_reference)(*reference_voxels[0])
    reference_y = (ctypes.c_int * num_reference)(*reference_voxels[1])
    reference_z = (ctypes.c_int * num_reference)(*reference_voxels[2])

    # setup return and call the function
    initial_indices = (ctypes.c_int * num_query)(*([-1] * num_query))
    VOXEL_C.get_neighbor_voxels.restype = ctypes.POINTER(ctypes.c_int * num_query)
    first_shell_indices_c = VOXEL_C.get_neighbor_voxels(
        ctypes.byref(query_x),
        ctypes.byref(query_y),
        ctypes.byref(query_z),
        ctypes.byref(reference_x),
        ctypes.byref(reference_y),
        ctypes.byref(reference_z),
        num_query,
        num_reference,
        ctypes.byref(initial_indices),
    )
    first_shell_indices = [i for i in first_shell_indices_c.contents if i != -1]

    return (
        np.array([query_voxels[0][i] for i in first_shell_indices]),
        np.array([query_voxels[1][i] for i in first_shell_indices]),
        np.array([query_voxels[2][i] for i in first_shell_indices]),
    )


def get_first_shell_exposed_voxels(
    exposed_voxels: VoxelGroup,
    buried_voxels: VoxelGroup,
    voxel_grid: VoxelGrid,
    performant: bool = utils.using_performant(),
) -> VoxelGroup:
    """
    Subset exposed voxels into only those neighboring one or more buried voxels.
    """
    if performant:
        first_shell_voxels = get_neighbor_voxels_c(
            exposed_voxels.voxels, buried_voxels.voxels
        )
    else:
        first_shell_voxels = get_neighbor_voxels_python(
            exposed_voxels.voxels, buried_voxels.voxels
        )

    first_shell_indices = compute_voxel_indices(first_shell_voxels, voxel_grid.x_y_z)

    return VoxelGroup(
        voxels=(
            np.array(first_shell_voxels[0]),
            np.array(first_shell_voxels[1]),
            np.array(first_shell_voxels[2]),
        ),
        indices=first_shell_indices,
        num_voxels=len(first_shell_indices),
        voxel_type="exposed",
        volume=compute_voxel_group_volume(len(first_shell_indices)),
    )


def is_neighbor_voxel(
    voxel_one: tuple[np.int64, ...], voxel_two: tuple[np.int64, ...]
) -> bool:
    """
    Given two voxels return True if they are ordinal neighbors.
    """
    # below approach is ~2x faster than using a list comprehension, presumably because of early exit
    sum_differences = 0
    for dimension in range(3):
        difference = abs(voxel_one[dimension] - voxel_two[dimension])

        if difference > 1:
            return False

        sum_differences += difference

    # a voxel is an ordinal neighbor when the sum of the absolute differences in axes indices is 1
    if sum_differences == 1:
        return True

    return False


def get_single_voxel(
    voxels: tuple[np.ndarray, ...], index: int
) -> tuple[np.int64, ...]:
    """
    Given a set of voxels return just one voxel at index.
    """
    return (
        voxels[0][index],
        voxels[1][index],
        voxels[2][index],
    )


def compute_voxel_indices(
    voxels: tuple[np.ndarray, ...], grid_dimensions: np.ndarray
) -> set[int]:
    """
    Given a 3D array of voxels, compute the 1D set of their indices.
    """
    grid_dimension_1_2 = grid_dimensions[1] * grid_dimensions[2]

    return {
        (
            voxels[0][i] * grid_dimension_1_2
            + voxels[1][i] * grid_dimensions[2]
            + voxels[2][i]
        )
        for i in range(len(voxels[0]))
    }


def compute_voxel_group_volume(num_voxels: int) -> float:
    """
    Return the volume of a number of voxels
    """
    return num_voxels * utils.VOXEL_VOLUME


def breadth_first_search_python(
    voxels: tuple[np.ndarray, ...], searchable_indices: set[int]
) -> set[int]:
    """
    Given a set of voxels and list of possible indices to add,
    add indices for all ordinal neighbors iteratively until no more such neighbors exist.
    """
    searchable_indices = deepcopy(searchable_indices)
    start_index = searchable_indices.pop()
    queue_indices = set([start_index])
    neighbor_indices = set([start_index])

    while len(queue_indices) > 0:
        current_index = queue_indices.pop()
        current_voxel = get_single_voxel(voxels, current_index)
        for searched_index in searchable_indices:
            searched_voxel = get_single_voxel(voxels, searched_index)
            if is_neighbor_voxel(current_voxel, searched_voxel):
                queue_indices.add(searched_index)
                neighbor_indices.add(searched_index)
        searchable_indices -= neighbor_indices

    return neighbor_indices


def breadth_first_search_c(
    voxels: tuple[np.ndarray, ...], searchable_indices: set[int]
) -> set[int]:
    """
    Given a set of voxels and list of possible indices to add,
    add indices for all ordinal neighbors iteratively until no more such neighbors exist.
    """
    c_searchable_indices = list(deepcopy(searchable_indices))

    # make the voxels compatible with C
    num_voxels = len(voxels[0])
    voxels_x = (ctypes.c_int * num_voxels)(*voxels[0])
    voxels_y = (ctypes.c_int * num_voxels)(*voxels[1])
    voxels_z = (ctypes.c_int * num_voxels)(*voxels[2])

    # generate other inputs needed for C function
    c_searchable_indices = (ctypes.c_int * len(c_searchable_indices))(
        *c_searchable_indices
    )
    initial_indices = (ctypes.c_int * num_voxels)(*([-1] * num_voxels))

    # run the breadth-first search
    VOXEL_C.breadth_first_search.restype = ctypes.POINTER(ctypes.c_int * num_voxels)
    neighbor_indices_c = VOXEL_C.breadth_first_search(
        ctypes.byref(voxels_x),
        ctypes.byref(voxels_y),
        ctypes.byref(voxels_z),
        ctypes.byref(c_searchable_indices),
        len(searchable_indices),
        ctypes.byref(initial_indices),
    )

    return set([i for i in neighbor_indices_c.contents if i != -1])


def breadth_first_search(
    voxels: tuple[np.ndarray, ...],
    searchable_indices: set[int],
    performant: bool = utils.using_performant(),
) -> set[int]:
    """
    Given a set of voxels and list of possible indices to add,
    add indices for all ordinal neighbors iteratively until no more such neighbors exist.
    """
    if performant:
        return breadth_first_search_c(voxels, searchable_indices)

    return breadth_first_search_python(voxels, searchable_indices)


def is_edge_voxel(voxel: np.ndarray, voxel_grid_dimensions: np.ndarray) -> bool:
    """
    Is this voxel at the edge of our grid on any dimension?
    """
    if 0 in voxel:
        return True
    elif voxel[0] == voxel_grid_dimensions[0] - 1:
        return True
    elif voxel[1] == voxel_grid_dimensions[1] - 1:
        return True
    elif voxel[2] == voxel_grid_dimensions[2] - 1:
        return True

    return False


def get_agglomerated_type(
    query_indices: set[int],
    buried_voxels: tuple[np.ndarray, ...],
    exposed_voxels: tuple[np.ndarray, ...],
    voxel_grid_dimensions: np.ndarray,
) -> tuple[set[int], str]:
    """
    Find "surface" voxels, being buried voxel in direct contact with an exposed voxel. Four possibilites:

    a) there are no "surface" voxels -> this is a cavity
    b) all "surface" voxels can be agglomerated (BFS) into a single group -> this is a pocket
    c) the "surface" voxels can be agglomerated (BFS) into exactly two groups -> this is a pore
    d) the "surface" voxels cannot be agglomerated (BFS) into less than 3 groups -> this is a hub
    """
    direct_surface_indices = set()
    for query_index in query_indices:
        query_voxel = get_single_voxel(buried_voxels, query_index)
        if is_edge_voxel(query_voxel, voxel_grid_dimensions):
            direct_surface_indices.add(query_index)
            continue
        for exposed_voxel_index in range(len(exposed_voxels[0])):
            exposed_voxel = get_single_voxel(exposed_voxels, exposed_voxel_index)
            if is_neighbor_voxel(query_voxel, exposed_voxel):
                direct_surface_indices.add(query_index)
                break

    # if there are no surface contacts, this must be a cavity
    if len(direct_surface_indices) == 0:
        return direct_surface_indices, "cavity"

    # add all voxels that are neighbours to the direct surface voxels
    neighbor_surface_indices = set()
    for query_index in query_indices - direct_surface_indices:
        query_voxel = get_single_voxel(buried_voxels, query_index)
        for surface_index in direct_surface_indices:
            surface_voxel = get_single_voxel(buried_voxels, surface_index)
            if is_neighbor_voxel(surface_voxel, query_voxel):
                neighbor_surface_indices.add(query_index)
                break

    # we pass all surface indices and the buried voxels for BFS
    # If there is more than one distinct surface (e.g. a pore) then
    #   BFS will terminate after agglomerating just one of the surfaces
    #   and `single_surface_indices` will be less than `surface_indices`
    # NOTE: we have to union the direct and neighbor surfaces, otherwise small discritization
    #   on the surface would look like a distinct surface
    surface_indices = direct_surface_indices.union(neighbor_surface_indices)
    single_surface_indices = breadth_first_search(buried_voxels, surface_indices)
    if len(single_surface_indices) < len(surface_indices):
        # run the agglomeration one more time on what's left
        remaining_surface_indices = surface_indices - single_surface_indices
        single_surface_indices = breadth_first_search(
            buried_voxels, remaining_surface_indices
        )

        # if there are stil indices left, there were more than 2 surfaces, hence a hub
        if len(single_surface_indices) < len(remaining_surface_indices):
            return direct_surface_indices.union(neighbor_surface_indices), "hub"

        # othewise, exactly 2 and a pore
        return direct_surface_indices.union(neighbor_surface_indices), "pore"

    # if the first BFS went to completion and consumed all surfaces, there was only a single surface, hence pocket
    return direct_surface_indices.union(neighbor_surface_indices), "pocket"


def get_voxel_group_center(
    voxel_indices: set[int], voxel_grid: VoxelGrid
) -> np.ndarray:
    """
    Compute the center of geometry for the voxel group.
    """
    voxel_coords = np.array(
        [voxel_grid.voxel_centers[voxel_index] for voxel_index in voxel_indices]
    )
    return np.mean(voxel_coords, axis=0)


def get_voxel_group_axial_lengths(
    voxel_indices: set[int], voxel_grid: VoxelGrid
) -> list[float]:
    """
    Align the voxel group to it's principal axes, then compute the maximum length along each axis.
    """
    voxel_coords = np.array(
        [voxel_grid.voxel_centers[voxel_index] for voxel_index in voxel_indices]
    )
    if len(voxel_coords) < 3:
        return [0, 0, 0]

    # rotation, translation = align.get_principal_axis_alignment_translation_rotation(voxel_coords)
    voxel_coords = align.align_structure(voxel_coords)

    # compute the maximum length across each axis
    xs = [voxel_coord[0] for voxel_coord in voxel_coords]
    ys = [voxel_coord[1] for voxel_coord in voxel_coords]
    zs = [voxel_coord[2] for voxel_coord in voxel_coords]

    return sorted(
        [
            round(max(xs) - min(xs), 3),
            round(max(ys) - min(ys), 3),
            round(max(zs) - min(zs), 3),
        ],
        reverse=True,
    )


def get_pores_pockets_cavities_occluded(
    buried_voxels: VoxelGroup, exposed_voxels: VoxelGroup, voxel_grid: VoxelGrid
) -> tuple[
    dict[int, VoxelGroup],
    dict[int, VoxelGroup],
    dict[int, VoxelGroup],
    dict[int, VoxelGroup],
    dict[int, VoxelGroup],
]:
    """
    Agglomerate buried solvent voxels into hubs, pores, pockets, cavities, and simply occluded.
    """
    buried_indices = set(range(buried_voxels.voxels[0].size))
    agglomerated_indices = set()

    hubs, pores, pockets, cavities, occluded = {}, {}, {}, {}, {}
    hub_id, pore_id, pocket_id, cavity_id, occluded_id = 0, 0, 0, 0, 0

    while len(agglomerated_indices) < len(buried_indices):
        remaining_indices = buried_indices - agglomerated_indices

        # perform BFS over the remaining indices
        agglomerable_indices = breadth_first_search(
            buried_voxels.voxels, remaining_indices
        )
        # iterate our counter of finished indices
        agglomerated_indices = agglomerated_indices.union(agglomerable_indices)

        # if too small, don't assign direct surface indices and assign type "occluded"
        if len(agglomerable_indices) <= MIN_NUM_VOXELS:
            direct_surface_indices = set()
            agglomerated_type = "occluded"
        else:
            # identify what these agglomerated voxels are
            direct_surface_indices, agglomerated_type = get_agglomerated_type(
                agglomerable_indices,
                buried_voxels.voxels,
                exposed_voxels.voxels,
                voxel_grid.x_y_z,
            )

        # get the surface voxels for use in getting their voxel-grid indices
        surface_voxels = (
            np.array(
                [buried_voxels.voxels[0][index] for index in direct_surface_indices]
            ),
            np.array(
                [buried_voxels.voxels[1][index] for index in direct_surface_indices]
            ),
            np.array(
                [buried_voxels.voxels[2][index] for index in direct_surface_indices]
            ),
        )

        # get the voxels
        volume_voxels = (
            np.array(
                [buried_voxels.voxels[0][index] for index in agglomerable_indices]
            ),
            np.array(
                [buried_voxels.voxels[1][index] for index in agglomerable_indices]
            ),
            np.array(
                [buried_voxels.voxels[2][index] for index in agglomerable_indices]
            ),
        )
        # get the voxel indices
        volume_indices = compute_voxel_indices(volume_voxels, voxel_grid.x_y_z)
        # create the voxelgroup
        voxel_group = VoxelGroup(
            voxels=volume_voxels,
            indices=volume_indices,
            surface_indices=compute_voxel_indices(surface_voxels, voxel_grid.x_y_z),
            num_voxels=len(volume_indices),
            voxel_type=agglomerated_type,
            volume=compute_voxel_group_volume(len(volume_indices)),
            center=get_voxel_group_center(volume_indices, voxel_grid),
            axial_lengths=get_voxel_group_axial_lengths(volume_indices, voxel_grid),
        )

        # add the voxelgroup depending on type and increment the type counter
        if agglomerated_type == "hub":
            hubs[hub_id] = voxel_group
            hub_id += 1
        elif agglomerated_type == "pore":
            pores[pore_id] = voxel_group
            pore_id += 1
        elif agglomerated_type == "pocket":
            pockets[pocket_id] = voxel_group
            pocket_id += 1
        elif agglomerated_type == "cavity":
            cavities[cavity_id] = voxel_group
            cavity_id += 1
        elif agglomerated_type == "occluded":
            occluded[occluded_id] = voxel_group
            occluded_id += 1

    return hubs, pores, pockets, cavities, occluded



"""
Various utility functions
"""


from typing import Optional

import numpy as np
import pandas as pd

from volumizer import constants
from volumizer.types import Annotation, VoxelGroup
from volumizer.paths import C_CODE_DIR


VOXEL_SIZE = constants.VOXEL_SIZE
VOXEL_VOLUME = VOXEL_SIZE**3
KEEP_MODELS = (
    True  # by default keep all models in case they are other biological assembly units
)
KEEP_NON_PROTEIN = False  # by default only keep protein residues
KEEP_HYDROGENS = False  # by default remove hydrogens


def set_resolution(resolution: float) -> None:
    """
    Set the value of the VOXEL_SIZE and VOXEL_VOLUME global constants.
    """
    global VOXEL_SIZE
    global VOXEL_VOLUME

    VOXEL_SIZE = resolution
    VOXEL_VOLUME = VOXEL_SIZE**3


def set_non_protein(non_protein: bool) -> None:
    """
    Set whether to include non-protein residues during the calculations.
    """
    global KEEP_NON_PROTEIN
    KEEP_NON_PROTEIN = non_protein


def get_volume_summary(
    voxel_group_dict: dict[int, VoxelGroup], summary_type: str = "total"
) -> float:
    """
    Compute a summary value for the volume
    """
    volumes = [
        voxel_group.volume
        for voxel_group in voxel_group_dict.values()
        if voxel_group.volume is not None
    ]
    if volumes:
        if summary_type == "total":
            return sum(volumes)
        elif summary_type == "max":
            return max(volumes)
        elif summary_type == "mean":
            return np.mean(volumes)
        else:
            print("Unsupported volume summary type")

    return 0.0


def sort_voxelgroups_by_volume(
    voxelgroups: dict[int, VoxelGroup]
) -> dict[int, VoxelGroup]:
    """
    Take a dictionary with indices as the keys and VoxelGroups as the values.
    Reassign the keys such that the VoxelGroups are sorted by volume.
    """
    return {
        i: voxelgroup
        for i, voxelgroup in enumerate(
            sorted(voxelgroups.values(), key=lambda group: group.volume, reverse=True)
        )
    }


def filter_voxelgroups_by_volume(
    voxelgroups: dict[int, VoxelGroup],
    min_volume: Optional[float] = None,
    min_voxels: Optional[int] = None,
) -> dict[int, VoxelGroup]:
    """
    Remove voxel groups based on min volumem and min number of voxels cutoff
    """
    if min_volume is not None:
        voxelgroups = {
            i: voxelgroup
            for i, voxelgroup in enumerate(voxelgroups.values())
            if voxelgroup.volume >= min_volume
        }
    if min_voxels is not None:
        voxelgroups = {
            i: voxelgroup
            for i, voxelgroup in enumerate(voxelgroups.values())
            if voxelgroup.num_voxels >= min_voxels
        }

    return voxelgroups


def make_annotation_dataframe(annotation: Annotation) -> pd.DataFrame:
    """
    Return a dataframe of the annotation values for each volume type.
    """
    hubs_df = pd.DataFrame.from_dict(
        [
            {
                "id": i,
                "type": "hub",
                "volume": annotation.hub_volumes[i],
                "x": annotation.hub_dimensions[i][0],
                "y": annotation.hub_dimensions[i][1],
                "z": annotation.hub_dimensions[i][2],
            }
            for i in annotation.hub_volumes.keys()
        ]
    )

    pores_df = pd.DataFrame.from_dict(
        [
            {
                "id": i,
                "type": "pore",
                "volume": annotation.pore_volumes[i],
                "x": annotation.pore_dimensions[i][0],
                "y": annotation.pore_dimensions[i][1],
                "z": annotation.pore_dimensions[i][2],
            }
            for i in annotation.pore_volumes.keys()
        ]
    )

    cavities_df = pd.DataFrame.from_dict(
        [
            {
                "id": i,
                "type": "cavity",
                "volume": annotation.cavity_volumes[i],
                "x": annotation.cavity_dimensions[i][0],
                "y": annotation.cavity_dimensions[i][1],
                "z": annotation.cavity_dimensions[i][2],
            }
            for i in annotation.cavity_volumes.keys()
        ]
    )

    pockets_df = pd.DataFrame.from_dict(
        [
            {
                "id": i,
                "type": "pocket",
                "volume": annotation.pocket_volumes[i],
                "x": annotation.pocket_dimensions[i][0],
                "y": annotation.pocket_dimensions[i][1],
                "z": annotation.pocket_dimensions[i][2],
            }
            for i in annotation.pocket_volumes.keys()
        ]
    )

    return pd.concat([hubs_df, pores_df, cavities_df, pockets_df], ignore_index=True)


def using_performant() -> bool:
    """
    Return True if the performant .so libraries exist.
    """
    if (C_CODE_DIR / "voxel.so").is_file() and (C_CODE_DIR / "fib_sphere.so").is_file():
        return True

    return False



"""
Define project paths.
"""

from pathlib import Path


ROOT_DIR = Path(__file__).resolve().parents[1]

C_CODE_DIR = ROOT_DIR / "src"
PACKAGE_DIR = ROOT_DIR / "volumizer"
TEST_DIR = ROOT_DIR / "tests"



"""
Functions for creating points on a sphere surface.
"""


import ctypes

import pandas as pd
import numpy as np

from volumizer import utils
from volumizer.constants import ATOMIC_RADII, BASE_ATOMIC_RADIUS
from volumizer.paths import C_CODE_DIR


GOLDEN_RATIO = (1 + np.sqrt(5.0)) / 4

if utils.using_performant():
    FIB_SPHERE_C_PATH = C_CODE_DIR / "fib_sphere.so"
    FIB_SPHERE_C = ctypes.CDLL(str(FIB_SPHERE_C_PATH.absolute()))


def fibonacci_sphere(radius: float, x: float, y: float, z: float, samples: int) -> dict[str, list[float]]:
    """
    Generate `samples` points at `radius` from `x,y,z` such that points are roughly equally spaced.
    """
    # this function could likely be made more performant by using numpy more intelligently
    x_coords = []
    y_coords = []
    z_coords = []

    for i in range(samples):
        phi = np.arccos(1 - 2 * (i + 0.5) / samples)
        theta = (np.pi * i) / GOLDEN_RATIO

        x_coords.append(x + (np.cos(theta) * np.sin(phi)) * radius)
        y_coords.append(y + (np.sin(theta) * np.sin(phi)) * radius)
        z_coords.append(z + (np.cos(phi)) * radius)

    return {"x": x_coords, "y": y_coords, "z": z_coords}


def get_example_point_distance(coords: dict[str, list[float]]) -> float:
    """
    Get the distance between two adjacent points on a fibonacci sphere.
    Due to the way the spiral traverses the sphere, for an unknown number of points on the sphere,
    we can only be certain that the first and second, or last and second last points are "adjacent"
    """
    coord_diff = [
        coords["x"][0] - coords["x"][1],
        coords["y"][0] - coords["y"][1],
        coords["z"][0] - coords["z"][1],
    ]
    return np.linalg.norm(coord_diff)


def estimate_fibonacci_sphere_samples(radius: float, voxel_size: float) -> int:
    """
    Based on the voxel-size and radius being used, estimate how many samples will be needed.
    Smaller voxel-size requires more samples
    """
    num_samples = 20
    coords = fibonacci_sphere(radius, 0, 0, 0, num_samples)
    while get_example_point_distance(coords) > voxel_size:
        num_samples *= 2
        coords = fibonacci_sphere(radius, 0, 0, 0, num_samples)

    return num_samples


def get_fibonacci_sphere_radii(element: str, voxel_size: float) -> list[float]:
    """
    Based on the voxel-size being used, estimate the size and number of radii needed.
    Smaller voxel-size requires more radii.

    We always place a single radii at the VDW radius for the element corresponding to this
    atom, and only use additional, smaller radii when needed.
    """
    vdw_radius = ATOMIC_RADII.get(element, BASE_ATOMIC_RADIUS)
    radii = [vdw_radius]

    scale_factor = 1
    while min(radii) > voxel_size:
        radii.append(vdw_radius - (voxel_size * scale_factor))
        scale_factor += 1

    return radii


def add_extra_points_python(coords: pd.DataFrame, voxel_size: float = utils.VOXEL_SIZE) -> pd.DataFrame:
    """
    For each given point which represents the center of a heavy atom,
    add additional points on a surface around that point at one or more radii.
    """
    elemental_radii = {element: get_fibonacci_sphere_radii(element, voxel_size) for element in set(coords["element"])}

    extra_x, extra_y, extra_z, extra_element = [], [], [], []
    for element, radii in elemental_radii.items():
        for radius in radii:
            num_samples = estimate_fibonacci_sphere_samples(radius, voxel_size)
            for _, coord in coords[coords["element"] == element].iterrows():
                extra_points = fibonacci_sphere(radius, coord.x, coord.y, coord.z, num_samples)
                extra_x += extra_points["x"]
                extra_y += extra_points["y"]
                extra_z += extra_points["z"]
                extra_element += [element] * num_samples

    extra_coords = pd.DataFrame.from_dict({"x": extra_x, "y": extra_y, "z": extra_z, "element": extra_element})
    coords = pd.concat([coords, extra_coords], ignore_index=True)

    return coords


def add_extra_points_c(coords: pd.DataFrame, voxel_size: float = utils.VOXEL_SIZE) -> pd.DataFrame:
    """
    For each given point which represents the center of a heavy atom,
    add additional points on a surface around that point at one or more radii.
    """
    elemental_radii = {element: get_fibonacci_sphere_radii(element, voxel_size) for element in set(coords["element"])}

    extra_x, extra_y, extra_z, extra_element = [], [], [], []
    for element, radii in elemental_radii.items():
        for radius in radii:
            num_samples = estimate_fibonacci_sphere_samples(radius, voxel_size)
            num_floats = num_samples * 3
            initial_values = (ctypes.c_float * num_floats)(*([0.0] * num_floats))
            FIB_SPHERE_C.fibonacci_sphere.restype = ctypes.POINTER(ctypes.c_float * num_floats)
            for _, coord in coords[coords["element"] == element].iterrows():
                c_coords = FIB_SPHERE_C.fibonacci_sphere(
                    ctypes.c_float(radius),
                    ctypes.c_float(coord.x),
                    ctypes.c_float(coord.y),
                    ctypes.c_float(coord.z),
                    num_samples,
                    ctypes.byref(initial_values),
                )
                computed_coords_flat = [coord for coord in c_coords.contents]

                extra_x += computed_coords_flat[:num_samples]
                extra_y += computed_coords_flat[num_samples : num_samples * 2]
                extra_z += computed_coords_flat[num_samples * 2 :]
                extra_element += [element] * num_samples

    extra_coords = pd.DataFrame.from_dict({"x": extra_x, "y": extra_y, "z": extra_z, "element": extra_element})
    coords = pd.concat([coords, extra_coords])

    return coords


def add_extra_points(
    coords: pd.DataFrame, voxel_size: float = utils.VOXEL_SIZE, performant: bool = utils.using_performant()
) -> pd.DataFrame:
    """
    For each given point which represents the center of a heavy atom,
    add additional points on a surface around that point at one or more radii.
    """
    if performant:
        return add_extra_points_c(coords, voxel_size)

    return add_extra_points_python(coords, voxel_size)



"""
Definition of custom types.
"""


from typing import NamedTuple, Optional

import numpy as np


class ComponentData(NamedTuple):
    component_id: str
    component_type: str


class VoxelGroup(NamedTuple):
    voxels: tuple[np.ndarray, ...]
    indices: set[int]
    num_voxels: int
    surface_indices: set[int] = set()
    voxel_type: Optional[str] = None
    volume: float = 0.0
    center: Optional[np.ndarray] = None
    axial_lengths: list[float] = [0.0, 0.0, 0.0]


class Annotation(NamedTuple):
    total_hub_volume: float
    total_pore_volume: float
    total_cavity_volume: float
    total_pocket_volume: float
    largest_hub_volume: float
    largest_pore_volume: float
    largest_cavity_volume: float
    largest_pocket_volume: float
    num_hubs: int
    num_pores: int
    num_cavities: int
    num_pockets: int
    hub_volumes: dict[int, Optional[float]]
    pore_volumes: dict[int, Optional[float]]
    cavity_volumes: dict[int, Optional[float]]
    pocket_volumes: dict[int, Optional[float]]
    hub_dimensions: dict[int, list[float]]
    pore_dimensions: dict[int, list[float]]
    cavity_dimensions: dict[int, list[float]]
    pocket_dimensions: dict[int, list[float]]



"""
Module containing functions for alignment of coordinates, e.g. along principal axes.
"""


import biotite.structure as bts


def align_structure(
    structure: bts.AtomArray,
) -> bts.AtomArray:
    """
    Translate the center-of-geometry of a protein to [0,0,0] and align to the principal axes.
    """
    return bts.orient_principal_components(structure)



"""
Functions for parsing, cleaning, and modifying PDBs.
"""

from pathlib import Path

import biotite.structure as bts
from biotite.structure.io import load_structure as biotite_load_structure
from biotite.structure.io import mmtf, pdbx, pdb
from biotite import InvalidFileError
from pyntcloud.structures.voxelgrid import VoxelGrid
import pandas as pd
import numpy as np

from volumizer.constants import (
    VOXEL_TYPE_CHAIN_MAP,
    VOXEL_TYPE_ATOM_MAP,
    VOXEL_TYPE_ELEMENT_MAP,
)
from volumizer.types import VoxelGroup
from volumizer import utils


def load_structure(file_path: Path) -> bts.AtomArray:
    """
    Load various structure formats.
    """
    try:
        if file_path.suffix == ".pdb":
            file = pdb.PDBFile.read(file_path)
            assembly = pdb.get_assembly(file, model=1)
        elif file_path.suffix == ".cif":
            file = pdbx.PDBxFile.read(file_path)
            assembly = pdbx.get_assembly(file, model=1)
        elif file_path.suffix == ".mmtf":
            file = mmtf.MMTFFile.read(file_path)
            assembly = mmtf.get_assembly(file, model=1)
        else:
            return biotite_load_structure(file_path)
        return assembly
    except InvalidFileError:
        return biotite_load_structure(file_path)


def save_pdb_lines(pdb_lines: list[str], output: Path | str) -> None:
    """
    Save individual pdb lines to a file.
    """
    if isinstance(output, (Path, str)):
        with open(output, mode="w", encoding="utf-8") as out_file:
            out_file.writelines("%s\n" % line for line in pdb_lines)
    else:
        output.writelines("%s\n" % line for line in pdb_lines)


def clean_structure(structure: bts.AtomArray) -> bts.AtomArray:
    """
    Clean the AtomArray of a PDB based on selected preferences.
    """
    if isinstance(structure, bts.AtomArrayStack):
        if not utils.KEEP_MODELS:
            structure = structure[0]
        elif structure.stack_depth() > 1:
            new_structure = structure[0]
            for model in structure[1:]:
                new_structure += model
            structure = new_structure

    if not utils.KEEP_NON_PROTEIN:
        structure = structure[bts.filter_amino_acids(structure)]

    if not utils.KEEP_HYDROGENS:
        structure = structure[~np.isin(structure.element, ["H"])]

    return structure


def get_structure_coords(structure: bts.AtomArray) -> pd.DataFrame:
    """
    Return the coordinates of an atom array as a dataframe.

    This is done because the PyntCloud package takes a dataframe as input to generate the cloud.
    """
    coordinates = structure.coord
    elements = structure.element

    coordinates = pd.DataFrame(coordinates, columns=["x", "y", "z"])
    coordinates["element"] = elements

    return coordinates


def structure_to_pdb(structure: bts.AtomArray) -> list[str]:
    """
    Convert an AtomArray into individual PDB lines
    """
    try:
        return [
            make_atom_line(
                atom.atom_name,
                atom.res_name,
                atom.res_id,
                atom.chain_id,
                idx,
                atom.coord,
                atom.element,
                beta=atom.b_factor,
            )
            for idx, atom in enumerate(structure)
        ]
    except AttributeError:
        return [
            make_atom_line(
                atom.atom_name,
                atom.res_name,
                atom.res_id,
                atom.chain_id,
                idx,
                atom.coord,
                atom.element,
            )
            for idx, atom in enumerate(structure)
        ]


def format_atom_name(atom_name: str) -> str:
    """
    Format an atom-name to occupy 4 characters the weird way PDBs do it.
    """
    if len(atom_name) == 1:
        return f" {atom_name}  "
    elif len(atom_name) == 2:
        return f" {atom_name} "
    elif len(atom_name) == 3:
        return f" {atom_name}"
    return atom_name[:4]


def make_atom_line(
    atom_name: str,
    res_name: str,
    res_num: int,
    chain_id: str,
    atom_index: int,
    coord: np.ndarray,
    element: str,
    beta: float = 0.0,
) -> str:
    """
    Make a single atom line in a PDB
    """
    return f"ATOM  {atom_index:>5d} {format_atom_name(atom_name)} {res_name} {chain_id}{res_num:>4d}    {coord[0]:>8.3f}{coord[1]:>8.3f}{coord[2]:>8.3f}  1.00{beta:>6.2f}           {element}"


def volume_to_structure(
    voxel_type: str,
    voxel_group_index: int,
    voxel_indices: set[int],
    surface_indices: set[int],
    voxel_grid_centers: np.ndarray,
) -> bts.AtomArray:
    """
    Convert one volume into a set of atoms in a biotite AtomArray
    """
    volume_structure = bts.AtomArray(len(voxel_indices))
    for i, voxel_index in enumerate(voxel_indices):
        if voxel_index in surface_indices:
            b_factor = 50.0
        else:
            b_factor = 0.0

        volume_structure[i] = bts.Atom(
            voxel_grid_centers[voxel_index], atom_id=voxel_index, b_factor=b_factor
        )

    volume_structure.atom_name = [VOXEL_TYPE_ATOM_MAP[voxel_type]] * len(voxel_indices)
    volume_structure.res_name = [voxel_type] * len(voxel_indices)
    volume_structure.res_id = [voxel_group_index] * len(voxel_indices)
    volume_structure.chain_id = [VOXEL_TYPE_CHAIN_MAP[voxel_type]] * len(voxel_indices)
    volume_structure.element = [VOXEL_TYPE_ELEMENT_MAP[voxel_type]] * len(voxel_indices)

    return volume_structure


def volumes_to_structure(
    voxel_grid: VoxelGrid,
    hubs: dict[int, VoxelGroup],
    pores: dict[int, VoxelGroup],
    pockets: dict[int, VoxelGroup],
    cavities: dict[int, VoxelGroup],
    occluded: dict[int, VoxelGroup],
) -> bts.AtomArray:
    """
    Convert the voxels of all volumes into a set atoms in a biotite AtomArray.
    """
    volume_structure = bts.AtomArray(0)

    for i, voxel_group in hubs.items():
        volume_structure += volume_to_structure(
            "HUB",
            i,
            voxel_group.indices,
            voxel_group.surface_indices,
            voxel_grid.voxel_centers,
        )
    for i, voxel_group in pores.items():
        volume_structure += volume_to_structure(
            "POR",
            i,
            voxel_group.indices,
            voxel_group.surface_indices,
            voxel_grid.voxel_centers,
        )
    for i, voxel_group in pockets.items():
        volume_structure += volume_to_structure(
            "POK",
            i,
            voxel_group.indices,
            voxel_group.surface_indices,
            voxel_grid.voxel_centers,
        )
    for i, voxel_group in cavities.items():
        volume_structure += volume_to_structure(
            "CAV",
            i,
            voxel_group.indices,
            voxel_group.surface_indices,
            voxel_grid.voxel_centers,
        )
    for i, voxel_group in occluded.items():
        volume_structure += volume_to_structure(
            "OCC",
            i,
            voxel_group.indices,
            voxel_group.surface_indices,
            voxel_grid.voxel_centers,
        )

    return volume_structure


def make_volumized_pdb_lines(
    structures: list[bts.AtomArray],
    deliminator: str = "END",
) -> list[str]:
    """
    Convenience function to produce the final set of PDB lines for the annotated PDB.
    """
    pdb_lines = []
    for structure in structures:
        pdb_lines.extend(structure_to_pdb(structure))
        pdb_lines.append(deliminator)

    return pdb_lines


def coordinates_to_structure(
    coords: np.ndarray, res_num: int = 0, chain_id: str = "A", element: str = "C"
) -> bts.AtomArray:
    """
    Convert a set of 3D coordinates into a biotite structure.

    By default these are carbon atoms.
    """
    # TODO: this needs testing alone and with pipeline
    #   left here for future improvement/use
    coordinate_structure = bts.AtomArray(len(coords))
    for i, coord in enumerate(coords):
        coordinate_structure[i] = bts.Atom(coord, atom_id=i, b_factor=0.0)

    coordinate_structure.atom_name = [element] * len(coords)
    coordinate_structure.res_name = ["CRD"] * len(coords)
    coordinate_structure.res_num = [res_num] * len(coords)
    coordinate_structure.chain_id = [chain_id] * len(coords)
    coordinate_structure.element = [element] * len(coords)

    return coordinate_structure



"""
Entry-level functions to find pores and cavities.
"""


from pathlib import Path
from typing import Optional

import biotite.structure as bts
import pandas as pd

from volumizer import utils, pdb, voxel, fib_sphere, align
from volumizer.types import Annotation


def annotate_structure_volumes(
    structure: bts.AtomArray,
    min_voxels: Optional[int] = 2,
    min_volume: Optional[float] = None,
) -> tuple[Annotation, list[str]]:
    """
    Perform analysis of a prepared structure.
    """
    coords = pdb.get_structure_coords(structure)
    coords = fib_sphere.add_extra_points(coords, utils.VOXEL_SIZE)

    cloud = voxel.coords_to_point_cloud(coords)
    cloud, voxel_grid_id = voxel.add_voxel_grid(cloud)
    voxel_grid = voxel.get_voxel_grid(cloud, voxel_grid_id)

    protein_solvent_voxels = voxel.get_protein_solvent_voxel_array(voxel_grid)
    protein_voxels, solvent_voxels = voxel.get_protein_and_solvent_voxels(
        protein_solvent_voxels, voxel_grid.x_y_z
    )

    exposed_voxels, buried_voxels = voxel.get_exposed_and_buried_voxels(
        solvent_voxels, protein_voxels, voxel_grid.x_y_z
    )

    # get sub-selection of exposed voxels that are NEXT to a buried voxel
    first_shell_exposed_voxels = voxel.get_first_shell_exposed_voxels(
        exposed_voxels, buried_voxels, voxel_grid
    )

    (
        hubs,
        pores,
        pockets,
        cavities,
        occluded,
    ) = voxel.get_pores_pockets_cavities_occluded(
        buried_voxels, first_shell_exposed_voxels, voxel_grid
    )
    hubs = utils.sort_voxelgroups_by_volume(
        utils.filter_voxelgroups_by_volume(
            hubs, min_voxels=min_voxels, min_volume=min_volume
        )
    )
    pores = utils.sort_voxelgroups_by_volume(
        utils.filter_voxelgroups_by_volume(
            pores, min_voxels=min_voxels, min_volume=min_volume
        )
    )
    pockets = utils.sort_voxelgroups_by_volume(
        utils.filter_voxelgroups_by_volume(
            pockets, min_voxels=min_voxels, min_volume=min_volume
        )
    )
    cavities = utils.sort_voxelgroups_by_volume(
        utils.filter_voxelgroups_by_volume(
            cavities, min_voxels=min_voxels, min_volume=min_volume
        )
    )
    occluded = utils.sort_voxelgroups_by_volume(
        utils.filter_voxelgroups_by_volume(
            occluded, min_voxels=min_voxels, min_volume=min_volume
        )
    )

    annotation = Annotation(
        total_hub_volume=utils.get_volume_summary(hubs, "total"),
        total_pore_volume=utils.get_volume_summary(pores, "total"),
        total_cavity_volume=utils.get_volume_summary(cavities, "total"),
        total_pocket_volume=utils.get_volume_summary(pockets, "total"),
        largest_hub_volume=utils.get_volume_summary(hubs, "max"),
        largest_pore_volume=utils.get_volume_summary(pores, "max"),
        largest_cavity_volume=utils.get_volume_summary(cavities, "max"),
        largest_pocket_volume=utils.get_volume_summary(pockets, "max"),
        num_hubs=len(hubs),
        num_pores=len(pores),
        num_cavities=len(cavities),
        num_pockets=len(pockets),
        hub_volumes={i: hub.volume for i, hub in hubs.items()},
        pore_volumes={i: pore.volume for i, pore in pores.items()},
        cavity_volumes={i: cavity.volume for i, cavity in cavities.items()},
        pocket_volumes={i: pocket.volume for i, pocket in pockets.items()},
        hub_dimensions={i: hub.axial_lengths for i, hub in hubs.items()},
        pore_dimensions={i: pore.axial_lengths for i, pore in pores.items()},
        cavity_dimensions={i: cavity.axial_lengths for i, cavity in cavities.items()},
        pocket_dimensions={i: pocket.axial_lengths for i, pocket in pockets.items()},
    )
    annotation_df = utils.make_annotation_dataframe(annotation)

    annotation_structure = pdb.volumes_to_structure(
        voxel_grid, hubs, pores, pockets, cavities, occluded
    )

    return (annotation_df, annotation_structure)


def prepare_pdb_structure(structure: bts.AtomArray) -> bts.AtomArray:
    """
    Prepares a biotite AtomArray object for analysis by:

    1. Cleaning extraneous atoms
    2. Aligning along the principal axis

    Returns a biotite AtomArray
    """
    cleaned_structure = pdb.clean_structure(structure)
    cleaned_aligned_structure = align.align_structure(cleaned_structure)

    return cleaned_aligned_structure


def volumize_structure(
    structure: bts.AtomArray,
) -> tuple[pd.DataFrame, bts.AtomArray, bts.AtomArray]:
    """
    Perform the volumizer pipeline on a biotite structure and return the:
    annotation_df, pdb_structure, and annotated_structure
    """
    prepared_structure = prepare_pdb_structure(structure)

    annotation_df, annotation_structure = annotate_structure_volumes(prepared_structure)

    return annotation_df, prepared_structure, annotation_structure


def volumize_pdb(pdb_file: Path) -> tuple[pd.DataFrame, bts.AtomArray, bts.AtomArray]:
    """
    Convenience function to volumize a PDB file.
    """
    pdb_structure = pdb.load_structure(pdb_file)
    return volumize_structure(pdb_structure)


def volumize_pdb_and_save(
    in_pdb_file: Path, out_pdb_file: Path, out_annotation_df: Path
) -> None:
    """
    Convenience function to perform volumization and save the output.
    """
    annotation_df, prepared_pdb_structure, annotation_structure = volumize_pdb(
        in_pdb_file
    )
    out_pdb_lines = pdb.make_volumized_pdb_lines(
        [prepared_pdb_structure, annotation_structure]
    )

    annotation_df.to_json(out_annotation_df)
    pdb.save_pdb_lines(out_pdb_lines, out_pdb_file)
